#ifndef GUI_H
#define GUI_H

#include <pthread.h>
#include "types.h"
#include "random.h"

// GUI's version of game state
#if DIM == 1
#include "history1d.h"
typedef struct str_gstate_t {
  u64b_t      time;     // Time step / generation (may be clipped)
  rnd_t       random;   // Random number generator state
  int         rolling;  // Are we animating (scrolling) 
  history_t  *history;  // History for class detection
  int         nhistory; // Max entries in history

  int        *rules;  // Rule table
  int         nrules; // Entries in rule table

  int         cellsx; // Width of 1D cell torus
  int         cellsy; // Visual history / height (y axis is time)
  int       **cells;

  int         r;      // Radius
  int         K;      // Number of states per cell

  int         c_type;
  int         c_cycle;
  int         c_stride;
	long        c_tick;
} gstate_t;
#elif DIM == 2
typedef struct str_gstate_t {
  u64b_t      time;     // Time step / generation (may be clipped)
  rnd_t       random;   // Random number generator state
  int         rolling;  // Are we animating (scrolling in time) 
  int         batch;    // Are we no GUI fast batch only?

  int         cellsx;   // Width of 1D cell torus
  int         cellsy;   // Visual history / height (y axis is time)
  int       **cells;
  int         w;
  int         h;


  float       hv;        // Constant bias
  int         r1;        // Radius1 (promotional)
  int         r2;        // Radius2 (inhibitory)
  float       j1;        // Scales radius 1 sum
  float       j2;        // Scales radium 2 sum

  int         converged; // 1 iff converged

  int          maxlen;
  int          characteristic_correlation_len;
  float       *spatial_correlation;
  float       *joint_entropy;
  float       *mutual_information;  
} gstate_t;
#endif

extern pthread_mutex_t StateLock;
extern pthread_mutex_t GuiLock;

////////////////////////////////////////////////////////////
// For GUI and widget files
////////////////////////////////////////////////////////////
#ifdef GUI_WIDGET
#include <X11/Xlib.h>
#include <GL/glx.h>
#include <GL/gl.h>

// Mouse button codes
#define MOUSE_LEFT   1
#define MOUSE_MIDDLE 2
#define MOUSE_RIGHT  3
#define MOUSE_UP     4
#define MOUSE_DOWN   5

// key and keycode
// These are probably machine + OS/xwindows specific
#define BACKSPACE 8   // Key
#define ESCAPE    27
#define ENTER     13
#define PAGEUP    99  // Keycode
#define PAGEDOWN  105
#define UP        98
#define DOWN      104
#define LEFT      100
#define RIGHT     102
#define F1        67
#define F2        68
#define F3        69
#define F4        70
#define F5        71
#define F6        72
#define F7        73
#define F8        74
#define F9        75
#define F10       76
#define F11       95
#define F12       96

// Holds what is needed to manage an opengl enabled window
typedef struct {
  Display             *dpy;
  int                  screen;
  Window               win;
  GLXContext           ctx;
  XSetWindowAttributes attr;
  int                  x, y;
  unsigned int         width,  height;
  unsigned int         pwidth, pheight;
  unsigned int         depth;   
  GLuint               font;
  int                  id;
} glwindow_t;

// Widget types and callbacks
typedef struct st_widget widget_t;

typedef void (*cbktick_t)      (widget_t *w);
typedef void (*cbkdraw_t)      (widget_t *w);
typedef void (*cbkkeypress_t)  (widget_t *w, char key, unsigned int keycode);
typedef void (*cbkmousedown_t) (widget_t *w, int x, int y, int b);
typedef void (*cbkmouseup_t)   (widget_t *w, int x, int y, int b);
typedef void (*cbkmousemove_t) (widget_t *w, int x, int y);

struct st_widget {
  glwindow_t     *glw;        // Window information
  float           x;          // x-pos
  float           y;          // y-pos
  float           w;          // width
  float           h;          // height
  cbkdraw_t       draw;       //
  cbkkeypress_t   keypress;   // 
  cbkmousedown_t  mousedown;  // Event callbacks
  cbkmouseup_t    mouseup;    //
  cbkmousemove_t  mousemove;  // 
  cbktick_t       tick;       // 
  int             update;     // Update request flag
  void           *wd;         // Custom widget data
};
#endif

////////////////////////////////////////////////////////////
// For files other than gui.c
////////////////////////////////////////////////////////////
#ifndef GUI_C

#ifdef ONEDCASIM_C
// These are intended for the simulation to start/update the gui
extern int  StartGUI(char *version, gstate_t *s); 
extern void UpdateGuiState(gstate_t *s);
#endif // !GUI_WIDGET

// These are intended for use by widgets / gui subcomponents
#ifdef GUI_WIDGET
extern gstate_t         *Stateg,*Statec,*Statep;

extern void   printGLf(GLuint font, const char *fmt, ...);
extern u32b_t LoadTexture(char *fn);
extern void   ViewPort3D(int x, int y, int w, int h);
extern void   ViewPort2D(glwindow_t *glw);

extern void AddWidget(glwindow_t *glw, float  x, float y, float w, float h,
		      cbkdraw_t draw, cbkkeypress_t keypress, cbkmousedown_t mousedown, 
		      cbkmouseup_t mouseup, cbkmousemove_t mousemove, cbktick_t tick, void *wd);

extern void Cyan();

extern void Yellow();
extern void Black();
extern void Red();
extern void Purple();
extern void Blue();
extern void White();
extern void Green();

extern float ScaleX(widget_t *w, const float x);
extern float ScaleY(widget_t *w, const float y);
extern void  GuiExit();

#endif // GUI_WIDGET
#endif // !GUI_C

#endif // !GUI_H
